''' Alapszintű helper függvények:  stingkezelés (pl. text2int, d_annotate, d_lookup), tömbök és szótárak kezelése, fájlrendszer, debug
''' 

from time import process_time,perf_counter
import os
from itertools import groupby
import string
import math
import numpy as np
import re


def stopperstart():
    return process_time()

def stopper(process_time_start):
    '''
    Előtte:  t=stopperstart()       stopper(t)     
    A nettó processzor-időt adja vissza sec-ban (sleep nélkül)
    Másik lehetőség:   perf_counter()   -   a teljes eltelt idő sec-ban (általában lényegesen nagyobb)
    '''
    print('process_time: ' + str(process_time()-process_time_start))


def fn_round(n, decimals=0):
    ''' Ez felel meg a standard kerekítési szabálynak (>=0.5 esetén felfelé egyébként lefelé; negatív esetén az abszolút értékre vonatkozik)
    A python round() függvénye bizonyos pontokon eltér ettől (pl. 2.5 esetén 2-t ad a kerekítés, más esetekben viszont rendben van).
    Ennek az a törekvés a magyarázata, hogy a python a lehető legkisebb eredő kerekeítési hibát próbálja meg elérni.
    '''
    multiplier = 10 ** decimals
    rounded_abs=np.floor(abs(n)*multiplier + 0.5) / multiplier
    return math.copysign(rounded_abs, n)



def kodto(kod,kodlista):
    ''' Szerializált dictionary, nem túl nagy kódlistákra 
    kodlista:  "kod:felirat//..."  
    '''
    if not kod: return ''
    if not kodlista: return ''
    if type(kod)!=str: kod=str(kod) 
    kereső=kod+':'
    nLen=len(kereső)
    aSor=kodlista.split('//')
    for sor in aSor:
        if sor[:nLen]==kereső: return sor[nLen:]
    return ''


def beginwith(strIn,samples,replace=None):   # -> found_sample vagy ''     replace esetén a cserével létrejövő string
    ''' samples:  minták | jeles felsorolása   példa:  'ki|be'
      FIGYELEM:  pont, zárójel és egyéb regex karakterek elé kötelezően '\' kell   (pl. r'\.')
    return:  üres vagy a talált minta
    process_time: mikrosec körül
    '''
    m=re.search(r'\A(' + samples + ')',strIn)
    if m==None: 
        if replace!=None: return strIn
        else: return ''
    else: 
        found=m.group()
        if replace!=None: return replace + strIn[len(found):]
        else: return found 


def endwith(strIn,samples,replace=None):   # -> found_sample vagy ''     replace esetén a cserével létrejövő string
    ''' samples:  minták | jeles felsorolása   példa:  'tól|től'
      FIGYELEM:  pont, zárójel és egyéb regex karakterek elé kötelezően '\' kell   (pl. r'\.')
    return:  üres vagy a talált minta
    process_time: mikrosec körül
    '''
    m=re.search('(' + samples + r')\Z',strIn)
    if m==None: 
        if replace!=None: return strIn
        else: return ''
    else: 
        found=m.group()
        if replace!=None: return strIn[:len(strIn)-len(found)] + replace
        else: return found 


def vanbenne(strIn,samples):
    ''' samples:  minták | jeles felsorolása   példa:  'dika|dike'
    return:  üres string vagy a talált minta  (ha több minta is jó, akkor a strIn-ben előrrébb álló minta)
    process_time: mikrosec körül
    '''
    m=re.search(r'(' + samples + ')',strIn)
    if m==None: return ''
    else: return m.group()

def cutleft(strIn,strCut):
    # Nem ellenőrzi, hogy valóban a strCut-tal kezdődik-e, egyszerűen levágja az elejéről a strCut hosszának megfelelő karaktereket
    return strIn[len(strCut):]

def trim(str):
    ''' strip a széleken + ismétlődő space-k törlése belül 
    Minden whitespace-re kiterjed  (a kimenetben csak space maradhat)
    '''
    return ' '.join(str.split())


def clean(str,accentlevel='soft',bDropSpaces=False):
    ''' lower, hosszú ékezet helyett rövid ékezet (a-á, e-é, o-ö, u-ü nincs összevonva), írásjelek törlése  (betűk és számjegyek maradnak)
    bDropSpaces:   True esetén a whitespace-ek teljes törlése, False esetén csak trim (a belső szóközök megmaradnak, az írásjelek helyett space)
    accentlevel: '': nincs összevonás,  'soft':  a-á, e-é, o-ö, u-ü nincs összevonva,    'hard':  erős összevonás
    '''
    if accentlevel=='hard':
        s1='áéíóöőúüű'
        s2='aeiooouuu'
    elif accentlevel=='soft':
        s1='íóőúű'
        s2='ioöuü'
    else:
        s1=''
        s2=''
    
    if bDropSpaces:
        str=str.translate(str.maketrans(s1,s2,string.punctuation + string.whitespace))
        return str.lower()
    else:
        str=str.translate(str.maketrans(string.punctuation,' '*len(string.punctuation)))   # írásjelek helyett szóközök
        str=str.translate(str.maketrans(s1,s2))
        str=' '.join(str.split())
        return str.lower()

def skipaccents(str,level='soft'):
    ''' Ékezetek elhagyása
    level:
       'soft':  csak a hosszú ékezeteket cseréli rövidre,  a-á, e-á, o-ö, u-ü eltérés megmarad
       'hard':  minden ékezetet elhagy
    '''
    if level=='soft':
        return str.translate(str.maketrans('íóőúű','ioöuü'))
    elif level=='hard':
        return str.translate(str.maketrans('áéíóöőúüű','aeiooouuu'))
    else:
        return str


def splitfirst(str,separator=None):
    '''  Hívása:  first,second = splitfirst(str,',') 
    Ha nincs benn határolójel, akkor a második változóba '' értéket ad vissza
    '''
    array=str.split(separator,1)
    if len(array)>1: return array[0],array[1]
    elif len(array)==1: return array[0],''
    else: return '',''

def text2int(text,bSerialOk=True,bCleaned=False,tupleout=False):
    ''' 
    text:  számjegyek, szövegesen kiírt szám vagy sorszám, római szám ('nulla', 'százhuszonöt',  max 999 billió)
       Példa:  "123"  "MCMLII", "háromezertizenkettő", "százhamincadik", "tizenkettes"  
       Általában egyetlen szó, de állhat több szóból is.
       case-insensitive, accent-insensitive, whitespace és írásjel érdektelen, szöveges változatnál szótövesít
       Érdemi elgépelések nem megengedettek (nem fuzzy jellegű)
       Nem lehet benne idegen szó (csak számjegyek, számszavak, római szám karakterek, ragok)
    bSorszamOk:  megengedett-e sorszám is  (az out ebben az esetben is egy szám lesz)
    bCleaned:   előzetes szabványosítás megtörtént-e  (egyetlen szó, lower, ékezet-összevonás).  30-40% gyorsulás érhető el 
    tupleout:  True esetén   (n,type)  a result, ahol type = 'szám', 'rómaiszám', 'sorszám'

    Result:  0-999 billió   -1 ha nem ismerhető fel számként
    process_time: 15 microsec körül   (ebből a fele idő a clean)

    '''


    def sub(nOut,tipus=''):
        if tupleout: return nOut,tipus
        else: return nOut


    if text=='': return sub(-1)

    # balról mindenképpen el kell tüntetni a whitespace-eket
    if not bCleaned:
        text=text.lstrip()
        if text=='': return sub(-1)

    # ha számjeggyel vagy '-' jellel kezdődik
    c=text[0]
    if c=='-' or c in string.digits:
        text=text.rstrip('. ')
        try:
            nOut=int(text)
            return sub(nOut,'szám')
        except:
            return sub(-1)


    if not bCleaned:
        text=clean(text,'hard',True)        # accent hard, ne maradjon whitespace   (viszonylag időigényes, kb 5 mikrosec)
        if text=='': return sub(-1)

    # Próbálja meg római számként értelmezni
    nOut=romaiszam2int(text)
    if nOut>0: return sub(nOut,'rómaiszám')


    # Ha sorszám is megengedett
    typeout='szám'

    if bSerialOk and ('dik' in text or 'els' in text or endwith(text,'s|stol|sig|son|sen|sban|sben')):
        # lemmatizálás  (nem minden rag, elsősorban a dátumokban előforduló számokra van kiélezve)
        if 'dik' in text: text=endwith(text,'dika|dike|diki|dikai|dikei|dikan|diken|dikaig|dikeig|dikatol|diketol|dikos|dikes|dikas','dik')
        elif 'els' in text: text=endwith(text,'elsotol|elson|elsoig|elseje|elsejen|elsejei|elsejeig|elsejetol','elso')
        else: text=endwith(text,'stol|sig|son|sen|sban|sben','s')
        
        d={ 'elso':'egy','egyedik':'egy','egyes':'egy','masodik':'ketto','kettedik':'ketto','kettes':'ketto',
            'harmadik':'harom','harmas':'harom','negyedik':'negy','negyes':'negy',
            'otodik':'ot','otos':'ot','hatodik':'hat','hatos':'hat','hetedik':'het','hetes':'het',
              'nyolcadik':'nyolc','nyolcas':'nyolc','kilencedik':'kilenc','kilences':'kilenc',
            'tizedik':'tiz','tizes':'tiz','huszadik':'husz','huszas':'husz','harmincadik':'harminc','harmincas':'harminc',
              'negvenedik':'negyven','negyvenes':'negyven','otvenedik':'otven','otvenes':'otven',
              'hatvanadik':'hatvan','hatvanas':'hatvan','hetvenedik':'hetven','hetvenes':'hetven',
              'nyolcvanadik':'nyolcvan','nyolcvanas':'nyolcvan','kilencvenedik':'kilencven','kilencvenes':'kilencven',
            'szazadik':'szaz','szazas':'szaz','ezredik':'ezer','ezres':'ezer','milliomodik':'millio','millios':'millio',
              'milliardodik':'milliard','milliardos':'milliard','billiomodik':'billio','billios':'billio'}
        for key,value in d.items():
            if text.endswith(key): 
                text=text[:-len(key)]+value
                typeout='sorszám'
                break


    nOut=0
    if text in ['null','nulla','zero']: return sub(0,'szám')

    numwords=[('billio',1000000000000),('milliard',1000000000),('millio',1000000),('ezer',1000),('szaz',100)]
    for numword,numvalue in numwords:
        nPos=text.find(numword)
        if nPos>-1:
            # előtte 1-999 szám állhat (kivéve a "száz" előtt: 1-9)
            nelottemax=999
            if numvalue==100: nelottemax=9

            elotte=text[:nPos]
            if elotte=='': nelotte=1
            else: nelotte=text2int(elotte)          # rekurzív hívás

            if nelotte<1 or nelotte>nelottemax: return sub(-1)
            nOut+=nelotte*numvalue
            text=text[nPos+len(numword):]       # a felhasznált rész levágása

    if text:
        dTizesek={'tizen':10,'tiz':10,'huszon':20,'husz':20,'harminc':30,'negyven':40,'otven':50,'hatvan':60,'hetven':70,'nyolcvan':80,'kilencven':90}    
        key,out = d_lookup(text,dTizesek,True)
        if out: 
            nOut+=out
            text=text[len(key):]      # a felhasznált rész levágása


    if text:
        dEgyesek={'egy':1,'ketto':2,'ket':2,'harom':3,'negy':4,'ot':5,'hat':6,'het':7,'nyolc':8,'kilenc':9}
        nEgyesek=dEgyesek.get(text)
        if nEgyesek==None: return sub(-1)
        nOut+=nEgyesek

    if nOut==0: return sub(-1)
    
    return sub(nOut,typeout)


def romaiszam2int(text):
    ''' Ha nem római szám, akkor 0
    kisbetűk is megengedettek'''
    text=text.lower()

    roman_numerals = {'i':1, 'v':5, 'x':10, 'l':50, 'c':100, 'd':500, 'm':1000}
    result = 0
    try:
        for i,c in enumerate(text):
            if (i+1) == len(text) or roman_numerals[c] >= roman_numerals[text[i+1]]:
                result += roman_numerals[c]
            else:
                result -= roman_numerals[c]
    except:
        result=0

    return result


def fileexists(path):
    return os.path.exists(path)


def unzip(aRecord,nCol=None):
    # Általános függvény
    #   aX,aY=unzip(aXy)
    #   aZ=unzip(aXyz,2)        - a nCol 0-bázisú
    # A visszaadott tömbök valójában tuple-ök. Ha a tömbelemek módosítására van szükség, akkor a list(aX) művelettel át kell térni listára.
    # Inverz művelet:   aRec=list(zip(aX,aY))
    if nCol: return list(zip(*aRecord))[nCol]
    else: return zip(*aRecord)

def sortrecords(aRec,byindex,reverse=False,outindex=None):
    # inplace rendezés a byindex sorszámú oszlop szerint
    # outindex:  visszaadja a rendezett aRec megadott sorszámú oszlopát (list)
    # FIGYELEM:  az np.nan értékekre rosszul rendez (ne legyen nan a rendezési oszlopban)
    #  - stringoszlop esetén az '' rendezési érték a végére kerül (itt nincs gond)
    aRec.sort(key = lambda x: x[byindex],reverse=reverse)
    if outindex!=None: return list(zip(*aRec))[outindex]

def sortarrays(aSorter,aToSort,reverse=False):
    # Az aSorter értékei alapján rendezi mindkét tömböt (inplace jellegű)
    aRec=list(zip(aSorter,aToSort))
    aRec.sort(key = lambda x: x[0],reverse=reverse)
    aSorter2,aToSort2=unzip(aRec)
    for i in range(len(aSorter2)): aSorter[i]=aSorter2[i]
    for i in range(len(aToSort2)): aToSort[i]=aToSort2[i]


def grouprecords(aRec,keylambda):
    # Rekordok csoportosítása egy kulcs szerint
    # - a kulcs általában egy kategória-mező:   keylambda = lambda x:x[0]
    # - a kategóriák képezhetők egy vagy több folytonos értékkészletű mező alapján is
    #   Példa: egy datetime mező kategorizálása aszerint, hogy melyik évbe esik    keylambda = lambda x: x[2].year
    # Előfeltétel:  from itertools import groupby
    
    # aRec:  list of tuple
    # keylambda:  átlalában egy lambda függvény, ami minden rekordhoz hozzárendeli a rendezési kulcsértéket

    # return:  két-elemű iterátor
    #   for key, records in grouprecords(aRec,lambda x:x[0])
    #      print(key)
    #      print(list(records))                     - a records szintén egy iterátor
    #      for rec in records: print(rec)

    aRec.sort(key=keylambda)        
    return groupby(aRec,keylambda)


def ddef(**fields):
    ''' Egy dictionary megadása függvény-arg formátumban
    Előnye a normál szintaxishoz képest:  áttekinthetőbb és nem kell idézőjelezni az adatneveket (csak egy-szavas adatnevekre jó)
    Példa:
        dict1 = { 'suptitle':'Felső sor', 'title':'Alsó sor' }
        dict1 = ddef( suptitle='Felső sor', title='Alsó sor' )
    '''
    return dset({},**fields)

def dget(dict,fields):
    # Többmezős unpacking dictionary-re   (nincs ilyen közvetlen művelet)
    #      Példa:   a,b,c = dget(dict,'a,b,c')
    #  fields:  lista vagy felsorolásos string (vessző határolással)
    #      - ha hiányzik valamelyik mezőnév, akkor None kerül a változóba  (nincs hibaüzenet)
    #      - default érték itt nem kell, a dictionary get függvénye eleve tudja
    if type(fields)==str: fields=fields.split(',')
    if len(fields)==1: return dict.get(fields[0])           # ilyenkor ne tuple-t adjon vissza
    elif len(fields)>0: return tuple(map(dict.get,fields)) 

def dset(dict,**fields):
    '''
    fields:  fieldname1=value1,fieldname2=value2, ....    
      - megadható közvetlenül vagy egy dict objektumként
      - közvetlen megadás esetén a mezőnevekhez nem kell idézőjel   (több szavas mezőnevekre nem működik)
      - ha valamelyik mezőnév hiányzik, akkor automatikus felvétel
      - None értékadás nem törli automatikusan a mezőt (None érték íródik be)
    '''
    aField=list(fields.keys())
    aValue=list(fields.values())
    for i in range(len(aField)): 
        dict[aField[i]]=aValue[i]
    return dict

def dsetsoft(dict,**fields):
    '''
    Soft:
      - csak akkor módosít, ha nincs még ilyen mező a dict-ben vagy None az értéke.
      - None értéket semmiképpen nem ír be
    fields:  fieldname1=value1,fieldname2=value2, ....    
      - megadható közvetlenül vagy egy dict objektumként
      - közvetlen megadás esetén a mezőnevekhez nem kell idézőjel   (több szavas mezőnevekre csak egy módosító dict működik)
      - ha valamelyik mezőnév hiányzik, akkor automatikus felvétel
    '''
    aField=list(fields.keys())
    aValue=list(fields.values())
    for i in range(len(aField)):
        if aValue[i]==None or dget(dict,aField[i])!=None: continue
        dict[aField[i]]=aValue[i]


def d_addhard(dict,key,value):
    ''' Akkor is hozzáadja az értéket a dict-hez, ha van már ilyen key, de ilyenkor sorszámozza a kulcsot.
    return:  a felvett tétel kulcsa  (sorszámozásban térhet el az input key-től)
    Nagy tömegű sorszámozás esetén is viszonylag gyors, mert az eddigi lemgmagasabb szorszámú kulcsot binárisan keresi.
    '''
    
    if key=='': return ''
    # ellenőrzés: van-e már ilyen kulcs
    keyout=key
    if dict.get(keyout):
        i = 2

        # First do an exponential search
        while dict.get(key+str(i)):
            i = i * 2

        # Result lies somewhere in the interval (i/2..i]
        # We call this interval (a..b] and narrow it down until a + 1 = b
        a, b = (i // 2, i)
        while a + 1 < b:
            c = (a + b) // 2 # interval midpoint
            a, b = (c, b) if dict.get(key+str(c)) else (a, c)

        keyout=key+str(b)

    dict[keyout]=value

    return keyout


def d_lookup(strIn,d_samples,tupleout=False,bWholesamples=True):
    ''' Alapesetben beginwith minták, de a d_samples teljes-egyezéses mintákat is tartalmazhat
    Beginwith minták esetén a leghosszabb minta érvényesül (a strIn a talált mintával kezdődik)
    Ha van találat, akkor az adott kulcs-hoz tartozó value-t adja vissza (a tupleout-tal kérhető a kulcs,values pár is return-kétn)
    Ha nincs találat, akkor None
    strIn: általában egyetlen szó   (szavanként, vagy néhány szavas ablakokkal kell hívni a függvényt)
    d_samples:   {[keresőszó]:[value],...}         A keresőszó string, a value lehet szám is vagy akár egy sub dictironary
       - a keresőszavakban csak számok és betűk lehetnek
       - a keresőszavak végén lehet egy pont:  teljes-egyezéses minta   (ha nincs pont, akkor beginwith egyezés is jó)
       
    tupleout: True esetén return (key,value).    A key lehet több szavas  (az esetleges lezáró pont nélkül adja vissza a függvény)
    bWholesamples:  vannak-e teljes-egyezéses minták is a d_samples-ben (pont a keresőszó végén) 
    process_time:   2 microsec     50-es dict, átlagos szó (nem talál)     Feltehetően nagy dict-re is gyors (hash index)
    '''
    
    result=None

    tosearch=strIn

    if bWholesamples:
        tosearch=tosearch.replace(' ','.') + '.'
        
        #result=d_samples.get(tosearch + '.')
        #if result: 
        #    if tupleout: return tosearch,result
        #    else: return result

    # Keresés hosszrövidítésekel (a leghosszabb keresőminta érvényesül)
    while len(tosearch)>1:
        result=d_samples.get(tosearch)
        if result: 
            if tupleout: 
                tosearch=tosearch.replace('.',' ')
                if tosearch[-1]==' ': tosearch=tosearch[:-1]
                return tosearch,result
            else: return result
        tosearch=tosearch[:-1]

    #if result:
    #    if tupleout: return tosearch,result
    #    else: return result
    #else:
    if tupleout: return None,None
    else: return None


       

def d_annotate(strIn,lookup,bNum=False,max_words_in_samples=1,accentlevel='soft'):    # -> pattern,invalues,outvalues
    ''' Annotálás jellegű művelet. A szöveg mintázatát állítja elő.  
    A beazonosítható szövegrészek helyébe szögletes zárójelben az entity-nevét írja, az eredeti és a szabványosított értéket pedig 
        menti az invalues és outvalues tömbbe (a két tömb indexelése a pattern-be kerülő "[...]" helyettesítőjelekhez igazodik) 

    strIn:  általában több szavas kifejezés vagy mondat.    
        Nem kell előzetesen szabványosítani (a jelen függvényben: lower, trim, skipaccents, skippunctuations)
    lookup:   {'[startsamples]':'[entity],[value]', ...}
        - [startsamples]: több keresőminta is megadható | határolással.  Példa:  'múlt|elmúlt|előző|legutóbbi'
                A sorrend érdektelen, mindig a lehetséges leghosszabb találat érvényesül ("tavasszal" akkor is talál, ha a "tavasz" minta előrébb áll)
                A keresőminták több szavasak is lehetnek (trimmelve kell megadni)
                Ha egy keresőminta végén pont van, akkor csak wholeword találat megengedett  (egyébként beginwith találat is jó)
        - [entity]:  entity-név (pl. "évszak"). Kötelezően egyszavas  ('_' lehet benne)  
                Ugyanaz az entitás-név több startwords-höz is megadható.  Tetszőleges számú entitásnév lehet a lookup-ben.
                "none" entitást kell megadni, ha csak szaványosításról illetve szinonimák összevonásáról van szó
                  Példa:  "korábbi|előző":none,korábbi.     Ebben az esetben csere a szabványosított értékre, az outvalues-be nem kerül bele
                "stopword" entitás-névvel kell felsorolni az elhagyandó szavakat (pl. névelők)
        - [value]:  ez az érték kerül be az outvalues-be (az adott entity névvel). Szabványosított / kivonatolt érték
                Ha nincs megadva, akkor a talált startsample lesz a szabványosított érték
                Lehet szám, dátum, string is.  String esetén több szavas is lehet (trimmelve kell megadni)
    bNum:  True esetén a szöveges vagy numerikus szám-szavak annotálása "[szám]" entity-névvel (sorszámok és római számok is)
    max_words_in_samples:  maximum hány szavas minták vannak a lookup-ben
        - ha kisebb szám van megadva, akkor az ennél hosszabb minták nem fognak találatot adni 
    accentlevel:     '': nincs összevonás,  'soft':  a-á, e-é, o-ö, u-ü nincs összevonva     'hard':  erős összevonás
      

    return   pattern,invalues,outvalues
                pattern:    példa:  "[entity1] egyébszó [entity2]"
                invalues:   az annotált részek eredeti értéke  (a fenti pattern-példához 2-elemű lista tartozik)
                outvalues:  az annotált részek kimeneti értéke  (a fenti pattern-példához 2-elemű lista tartozik)
    '''

    invalues=[]
    outvalues=[]

    dropchars=string.punctuation.replace('.','')        # a '.' karakteren kívüli összes írásjel

    # key-felsorolások kibontása (ha nincs még kibontva;  a lookup egy globális objektum, amit egyszer kell csak kibontani)
    if not lookup.get('d_compiled'):
        dict_compiled={'d_compiled':True}     # a d_compiled adat helyből beállítva
        for key,value in lookup.items():
            subkeys=key.split('|')          # több keresőminta is felsorolható '|' határolással
            for subkey in subkeys: 
                if subkey: 
                    subkey=subkey.lower()
                    subkey=skipaccents(subkey,accentlevel)
                    # írásjelek helyett szóközök (kivéve '.')
                    #strL=strL.translate(str.maketrans(dropchars,' '*len(dropchars)))

                    subkey=subkey.replace(' ','.')       # a több szavas keresőmintákban "." kell a szóközök helyett (technikai oka van)
                    dict_compiled[subkey]=value
        lookup=dict_compiled
    

    dropchars=string.punctuation.replace('.','')        # a '.' karakteren kívüli összes írásjel

    strL=strIn
    # írásjelek helyett szóközök (kivéve '.')
    strL=strL.translate(str.maketrans(dropchars,' '*len(dropchars)))
    # pontok helyett pont + space
    strL=strL.replace('.','. ')
    
    strL=strL.lower()
    strL=skipaccents(strL,accentlevel)
    
    # trim
    words=strL.split()

    # magában álló pontok elhagyása (nem volt előtte alfanum, érdektelen)
    wordsout=[]
    for word in words:
        if word=='.': continue
        wordsout.append(word)
    words=wordsout
    
    # Keresés szavanként (vagy pár szavas csúszóablakkal) a mintákat tartalmazó dict-ben  (like: karakter-elhagyások a végéről)
    wordsout=[]
    nLen=len(words)
    i=0
    while i<nLen:
        if bNum:
            word=words[i]     # a számoknál csak egyszavas keresés van
            if word!='hét':   # a "hét" szó helyébe ne írjon számot (összetéveszthető a "hét" időhatározóval)
                              # Később még ellenőrzöm, hogy ha [időtartam] a következő szó is, akkor [szám]-ra módosuljon az értelmezése (pl. "hét nap")
                n=text2int(word,True,False)
                if n>-1:
                    invalues.append(word)
                    outvalues.append(n)
                    wordsout.append('[szám]')
                    i+=1
                    continue

        # max_words_in_samples szóra kell keresni (ha van még ennyi szó)
        i2 = i + max_words_in_samples
        if i2>nLen: i2=nLen
        tosearch=' '.join(words[i:i2])
        
        key,foundvalue=d_lookup(tosearch,lookup,True)
        if foundvalue:
            # hány szavas a találat (ennyi szó lesz lecserélve a helyettesítőjellel)
            nFoundWords=len(key.split())
            
            entity,value=splitfirst(foundvalue,',')
            if entity=='stopword': wordout=''
            else:
                if value=='': value=key
                if entity!='none': 
                    # Ha az előző szó is [időtartam] volt "hét" input-értékkel, akkor előző szó legyen számként annotálva (pl. "hét nap" "[szám] [időtartam]"
                    if entity=='időtartam' and len(wordsout)>0 and wordsout[-1]=='[időtartam]' and invalues[-1]=='hét':
                        wordsout[-1]='[szám]'
                        outvalues[-1]=7
                    if entity=='szám': value=int(value)
                    invalues.append(' '.join(words[i:i+nFoundWords]))
                    outvalues.append(value)
                    wordout='[' + entity + ']'
                # Ha a keresőmintához "none" lett megadva entitásként, akkor nem helyettesítőjellel, hanem a szabványosított értékkel
                #    kell lecserélni a találatot eredményező szövegrészt  (a value több szavas is lehet) 
                else:
                    wordout=value
            i += nFoundWords

        else: 
            # Ha nincs találat, akkor egyetlen szó kerül be a wordsout-ba (változtatás nélkül), majd lépés a következő szóra
            wordout=words[i]
            i+=1



        if wordout!='': wordsout.append(wordout)
    



    #for word in words:
    #    if word=='.': continue   # a . előtt nem volt alfanum (érdektelen)

    #    if bNum:
    #        if word!='hét':   # a "hét" szó helyébe ne írjon számot (összetéveszthető a "hét" időhatározóval)
    #                          # Később még ellenőrzöm, hogy ha [időtartam] a következő szó is, akkor [szám]-ra módosuljon az értelmezése (pl. "hét nap")
    #            n=text2int(word,True,False)
    #            if n>-1:
    #                invalues.append(word)
    #                outvalues.append(n)
    #                wordsout.append('[szám]')
    #                continue

    #    key,foundvalue=d_lookup(word,lookup,True)
    #    if foundvalue:
    #        entity,value=splitfirst(foundvalue,',')
    #        if entity=='stopword': wordout=''
    #        else:
    #            if value=='': value=key
    #            if entity!='none': 
    #                # Ha az előző szó is [időtartam] volt "hét" input-értékkel, akkor előző szó legyen számként annotálva (pl. "hét nap" "[szám] [időtartam]"
    #                if entity=='időtartam' and len(wordsout)>0 and wordsout[-1]=='[időtartam]' and invalues[-1]=='hét':
    #                    wordsout[-1]='[szám]'
    #                    outvalues[-1]=7
    #                if entity=='szám': value=int(value)
    #                invalues.append(word)
    #                outvalues.append(value)
    #                wordout='[' + entity + ']'
    #            else:
    #                wordout=value
    #    else: wordout=word



    #    if wordout!='': wordsout.append(wordout)
    
        
    pattern=' '.join(wordsout)
    return pattern,invalues,outvalues

   
    # B VÁLTOZAT:  mintánként scan a teljes szövegre
    #for regex,data in d_regex:
    #    value0=clean(regex,'',False)
    #    if regex.endswith('*'): regex=regex[:-1]
    #    else: regex=regex + r'\b'
    #    if regex.startswith('*'): regex=regex[1:]
    #    else: regex=r'\b' + regex
    #    regex=regex.replace('*','.*')

    #    entity,value=splitfirst(data,',')
    #    if value=='': value=value0

    #    re.sub(regex,entity,strIn)

    return



def stringadd(str,stradd,sep=','):
    ''' NE HASZNÁLD   Régi vágású stringadd, nem túl sok részstringre;  NINCS INPLACE VÁLTOZAT, lines = stringadd(lines,line,'\n') utasítás kell.
    Korábbi programkódok migrációjához lehet szükséges 
      (a régi módszer bizonyos esetekben kevesebb programsort igényelt, könnyebben át lehetett tekinteni, de kevésbé rugalmas és lassabb volt)
    A Python stringek esetén nem oldható meg az "inplace" (var argumentumos) értékadás   (bytearray objektummal sem) 
    A standard python megoldás összehasonlítása a régi típusú stringadd algoritmussal:
        lines=[]                                lines=''
        for ...:   lines.append(line)           for ...:   stringadd(lines,line,'\n')     // a python-ban csak a lines=stringadd(lines,line,'\n') működik
        '\n'.join(lines)                        lines
    '''
    if str: return str + sep + stradd 
    else: return stradd


def joinlines(lines):
    return '\n'.join(lines)         # régi vágású helper függvény




def FvNumformat(format,var=''):
    # str.format függvény paraméterezése.  
    # var: általában üres;  a diagram-jelölők formázásakor 'x' kell.
    if format[:3]=='{' + var + ':': return format                             # ha teljes formátum van megadva
    elif format==',': return '{' + var + ':,}'                                # ezres tagolás
    elif format[-2:]=='f%': return '{' + var + ':.' + format[:-2] + 'f}%'     # közvetlen százalék (bázis=100)
    else: return '{' + var + ':,.' + format + '}'                             # pl. '2f'  '3g'   '5e',  '0%'





''' NLP modul for hungarian date-entity recognition and translation to specific date values
NLP modul, magyar nyelvű dátum-kifejezések felismerése és lefordítása konkrét dátum-értékekre.
A legfontosabb függvény a modulban:  text2date( )    (a többi függvény helper jellegű)
Paraméterezési lehetőség:   lookup_text2dateG a modul végén (a leírást lásd a modul törzsrészében)
Részletes leírás:  https://github.com/COREtxt/hundate-core
''' 

from datetime import datetime,date,timedelta
from dateutil.relativedelta import relativedelta

try:
    from ezhelper import *
except:
    from .ezhelper import *     # ha csomagban van telepítve (pip install hundate)


def fn_today(format='.'):
    # date.today() is jó
    if format=='c10.' or format=='.':
        return datetime.now().strftime('%Y.%m.%d')
    elif format=='c10-' or format=='-':
        return datetime.now().strftime('%Y-%m-%d')
    elif format=='c8':
        return datetime.now().strftime('%Y%m%d')

def fn_dateadd(dt,nToAdd,unit='day'):
    if unit=='month':
        return dt + relativedelta(months=nToAdd)
    elif unit=='year':
        return dt.replace(year=dt.year+nToAdd)
    elif unit=='week':
        return dt + timedelta(weeks=nToAdd)
    elif unit=='day':
        return dt + timedelta(days=nToAdd)

def fn_daydiff(dt1,dt2):
    return (dt1-dt2) / timedelta(days=1)

def fn_monthlastday(dt):
    # A megadott nap hónapjának utolsó napja
    return fn_dateadd(fn_dateadd(dt,1,'month'),-1)

def fn_monday(dt='today'):
    # A megadott nap hetének hétfői napja
    if dt=='today': dt=date.today()
    return fn_dateadd(dt,-dt.weekday())         # 0 bázisú

def fn_parsedate(strDate,format='.'):
    # Alapformátumok lefordítása datetime-ra
    # format: a strDate input formátuma
    #        '.'  'c10.'  '-' 'c10-'   'c8'  
    # return:  datetime vagy None
    if format=='c10.' or format=='.':
        return datetime.strptime(strDate,'%Y.%m.%d')
    elif format=='c10-' or format=='-':
        return datetime.strptime(strDate,'%Y-%m-%d')
    elif format=='c8':
        return datetime.strptime(strDate,'%Y%m%d')


def text2date(text,dt0=None,tense='',outtype='first'):
    ''' Magyar nyelvű időmeghatározások lefordítása egy dátumra vagy dátum tartományra
    text:  általában több szavas
        A mondatban dátum-meghatározástól független szavak is lehetnek
    dt0:  relációs dátummeghatározások esetén a kiinduló dátum.
        Ha nincs megadva, akkor a mai nap.
    tense: 'future' / 'past'.  A nem egyértelmű időmeghatározások esetén jövőbeli vagy múltbeli dátumot preferáljon a függvény
        Ha üres, akkor az aktuális évben/hónapban/héten
    outtype:
      'first':    return =  '',   '2021.10.12',  '2021.12.10-2021.12.20'     Az első előforduló dátum vagy dátumtartomány.
      'first_tuple':  return = (datetime1, datetime2)    Az első dátum None, ha nem talált dátum-kifejezést. A második dátum csak időszak-kifejezése esetén nem None
      'first+':   ugyanaz mint a first, de a string végére beírja a pattern-t is   Példa: '2021.10.12   pattern: [szám] [hónapnév] [szám]
      'all':      '2021.10.12,2021.12.10-2021.12.20'
    '''
    
    def sub_sorszam(szám_in):
        return  szám_in and endwith(szám_in,r'\.|dik|elsö')    # "első" helyett "első" kell (soft accent)

    def sub_időtartam_to_időszak(időtartam,múltjövő,dt0):
        dtout=None
        dtout2=None
        
        szám=1
        if múltjövő=='múlt': szám=-1
        elif múltjövő=='most': szám=0
        
        if időtartam=='nap':
            dtout=fn_dateadd(dt0,szám)
        elif időtartam=='hét':
            dtMonday=fn_monday(dt0)             # az aktuális hét első napja
            dtout=fn_dateadd(dtMonday,7*szám)   # előző, aktuális, következő
            dtout2=fn_dateadd(dtout,6)          
        elif időtartam=='hónap':
            dtMonthfirst=dt0.replace(day=1)
            dtout=fn_dateadd(dtMonthfirst,szám,'month')
            dtout2=fn_monthlastday(dtout)
        elif időtartam=='negyedév':
            dtFirst=dt0.replace(month=(1 + ((dt0.month-1)//3) * 3))
            dtFirst=dt0.replace(day=1)
            dtout=fn_dateadd(dtFirst,3*szám,'month')
            dtout2=fn_monthlastday(fn_dateadd(dtout,2,'month'))
        elif időtartam=='félév':
            dtFirst=dt0.replace(month=(1 + ((dt0.month-1)//6) * 6))
            dtFirst=dt0.replace(day=1)
            dtout=fn_dateadd(dtFirst,6*szám,'month')
            dtout2=fn_monthlastday(fn_dateadd(dtout,5,'month'))
        elif időtartam=='év':
            dtFirst=datetime(dt0.year,1,1)
            dtout=fn_dateadd(dtFirst,szám,'year')
            dtout2=datetime(dtout.year,12,31)
        elif időtartam=='évtized':
            dtFirst=datetime((dt0.year//10)*10,1,1)
            dtout=fn_dateadd(dtFirst,szám*10,'year')
            dtout2=datetime(dtout.year+9,12,31)
        elif időtartam=='évszázad':
            dtFirst=datetime((dt0.year//100)*100,1,1)
            dtout=fn_dateadd(dtFirst,szám*100,'year')
            dtout2=datetime(dtout.year+99,12,31)
        elif időtartam=='évezred':
            dtFirst=datetime((dty.year//1000)*1000,1,1)
            dtout=fn_dateadd(dtFirst,szám*1000,'year')
            dtout2=datetime(dtout.year+999,12,31)
        
        return (dtout,dtout2)

    def sub_húsvétvasárnap(év):   # return datetime vagy None
            honapnap=kodto(év,'2000:04.23//2001:04.15//2002:03.31//2003:04.20//2004:04.11//2005:03.27//2006:04.16//2007:04.08//' +
                    '2008:03.23//2009:04.12//2010:04.04//2011:04.24//2012:04.08//2013:03.31//2014:04.20//2015:04.05//2016:03.27//' +
                    '2017:04.16//2018:04.01//2019:04.21//2020:04.12//2021:04.04//2022:04.17//2023:04.09//2024:03.31//2025:04.20//' +
                    '2026:04.05//2027:03.28//2028:04.16//2029:04.01//2030:04.21//2031:04.13//2032:03.28//2033:04.17//2034:04.09//' +
                    '2035:03.25//2036:04.13//2037:04.05//2038:04.25//2039:04.10//2040:04.01//2041:04.21//2042:04.06//2043:03.29//' +
                    '2044:04.17//2045:04.09//2046:03.25//2047:04.14//2048:04.05//2049:04.18//2050:04.10')
            if honapnap: 
                return fn_parsedate(str(év) + '.' + honapnap,'.')
    
    def sub_évnapja(év,évnapja):   # return (date1,date2)
        
        # "MMdd"
        if len(évnapja)==4 and évnapja.isdigit():
            try: return ( fn_parsedate(str(év) + évnapja,'c8'), None )
            except: pass 
        # "MMdd:MMdd"
        elif len(évnapja)==9 and évnapja[:4].isdigit() and évnapja[5:].isdigit():
            try: return ( fn_parsedate(str(év) + évnapja[:4],'c8'), fn_parsedate(str(év) + évnapja[5:],'c8') )
            except: pass
        # "húsvét%"
        elif beginwith(évnapja,'húsvét'):
            date1=sub_húsvétvasárnap(év)
            if date1:
                add=cutleft(évnapja,'húsvét')
                if add=='_': return (fn_dateadd(date1,-2), fn_dateadd(date1,1) )    # péntektől hétfőig
                else:
                    try: nAdd=int(add)
                    except: nAdd=0
                    date1=fn_dateadd(date1,nAdd)
                    return ( date1, None )
                     
        # "pünkösd%"
        elif beginwith(évnapja,'pünkösd'):
            date1=sub_húsvétvasárnap(év)
            if date1:
                date1 = fn_dateadd(date1,49)        # 7 héttel húsvét után
                add=cutleft(évnapja,'pünkösd')
                if add=='_': return (date1, fn_dateadd(date1,1) )    # vasárnaptól hétfőig
                else:
                    try: nAdd=int(add)
                    except: nAdd=0
                    date1=fn_dateadd(date1,nAdd)
                    return ( date1, None )

        
        return (None,None)


    # Generalizált mintázatok
    def sub_patterns(patternL,invaluesL,outvaluesL,dt0,patternword_előtte,patternword_utána):   # -> (date1, date2)  vagy None 
        # patternL:  szögletes zárójeles helyettesítőjelek (annotált szavak) és egyéb szavak lehetnek benne
        #     - a lehetséges helyettesítőjeleket a lookup_text2dateG tartalmazza (továbbá: [dátum])
        #     - a helyettesítőjelek által képviselt eredeti érték több szavas is lehet (lásd invaluesL)
        # invaluesL:  az annotált szavak input értéke  (az elemei a patternL annotált szavaihoz igazodnak) 
        #    FIGYELEM  a magánhangzók rövid ékezettel kerülnek bele (soft accent)
        # outvaluesL: az annotált szavak (kifejezések) szabványosított értéke
        #    - itt lehetnek hosszú ékezetek

        if dt0==None: dt0=datetime.now()

        dtout=None
        dtout2=None

        # első körös mintázatok

        if patternL=='[szám] [szám] [szám]':
            n1,n2,n3=outvaluesL
            if (n1>=1000 and n1<=2100) and (n2>=1 and n2<=12) and (n3>=1 and n2<=31): 
                try: dtout=datetime(n1,n2,n3)
                except: dtout=None

        elif patternL=='[szám] [szám] [időtartam] [szám]':     # 2022. II.hó 4.
            n1,n2,időtartam,n3=outvaluesL
            if időtartam=='hónap':
                if (n1>=1000 and n1<=2100) and (n2>=1 and n2<=12) and (n3>=1 and n2<=31): 
                    szám_in=invaluesL[3]
                    if not endwith(szám_in,'dik|elsö'):
                        try: dtout=datetime(n1,n2,n3)
                        except: dtout=None

        elif patternL=='[szám] [hónapnév] [szám]':
            n1,honap,n2=outvaluesL
            # "első", "második", ... nem megengedett  (pl. "2022 március első péntek")
            szám_in=invaluesL[2]
            if not endwith(szám_in,'dik|elsö'):
                try: dtout=datetime(n1,int(honap),n2)
                except: dtout=None

        elif patternL=='[szám] [hónapnév]':
            n1,honap=outvaluesL
            n2=int(honap)
            if (n1>=1000 and n1<=2100) and (n2>=1 and n2<=12): 
                dtout=datetime(n1,n2,1)
                dtout2=fn_monthlastday(dtout)

        elif patternL=='[hónapnév] [szám]':
            honap,n2=outvaluesL
            szám_in = invaluesL[1]
            if not (endwith(szám_in,'dik') or szám_in=='elsö'):     # "január első fele" ne "január 1." legyen
                try: dtout=datetime(dt0.year,int(honap),n2)
                except: dtout=None

        elif patternL=='[szám] [szám]':
            n1,n2=outvaluesL
            # "2022 01"   
            if (n1>=1000 and n1<=2100) and (n2>=1 and n2<=12): 
                szám_in = invaluesL[1]
                if len(szám_in)==2 and szám_in.isdigit():    # fontos, hogy a "2021 2. ne kerüljön ide (pl. nem működne a "2021. második fele" beazonosítása)
                    dtout=datetime(n1,n2,1)
                    dtout2=fn_monthlastday(dtout)
            # "01 12"    hónap nap
            elif (n1>=1 and n1<=12) and (n2>=1 and n2<=31):
                try: dtout=datetime(dt0.year,n1,n2)
                except: dtout=None
            # "2012-2014"
            elif (n1>=1000 and n1<=2100) and (n2>=1000 and n2<=2100) and (n2>n1):
                dtout=datetime(n1,1,1)
                dtout2=datetime(n2,12,31)

        elif patternL=='[szám] [évszak]':   # "2015 tavasz"   
            n,évszak=outvaluesL         # évszak: 3,6,9,12

            if n>=1 and n<=2100:
                if évszak=='12': n-=1    # a tél átnyúlik az évhatáron, a kezdődátuma előző évi
                dtout=datetime(n,int(évszak),1)    
                dtout2=fn_monthlastday(fn_dateadd(dtout,2,'month'))

        elif patternL=='[szám]':
            n,=outvaluesL
            szám_in,=invaluesL
            # évszám
            if n>=1000 and n<=2100:
                dtout=datetime(n,1,1)
                dtout2=datetime(n,12,31)
            # másodikán, másodikától, ...  (de "első" "második" ne kerüljön ide)
            elif n>=1 and n<=31 and vanbenne(szám_in,'dika|dike|diká|diké|elsej'):
                try: dtout=datetime(dt0.year,dt0.month,n)
                except: dtout=None
            # '20220102'  '19551211'
            elif len(szám_in)==8:
                if beginwith(szám_in,'19|20'): 
                    try: dtout=datetime(int(szám_in[:4]),int(szám_in[4:6]),int(szám_in[6:]))
                    except: dtout=None
            # '0203'  '1201'
            elif len(szám_in)==4:
                if beginwith(szám_in,'0|1'): 
                    try: dtout=datetime(dt0.year,int(szám_in[:2]),int(szám_in[2:]))
                    except: dtout=None

        elif patternL=='[maholnap]':
            n,=outvaluesL
            dtout=fn_dateadd(dt0,int(n))       

        elif patternL=='[maholnap] [időtartam]':     # 'mai napon', 'holnapi napon', 
            maholnap,időtartam=outvaluesL
            n=int(maholnap)
            if időtartam=='nap':
                dtout=fn_dateadd(dt0,n)       

        elif patternL=='[maholnap] [utánelőtt]':     # 'holnap után', 'tegnap előtt', 
            maholnap,utánelőtt=outvaluesL
            n=int(maholnap)
            if n==1 and utánelőtt=='után': dtout=fn_dateadd(dt0,2)       
            elif n==-1 and utánelőtt=='előtt': dtout=fn_dateadd(dt0,-2)       

        elif (patternL=='[hónapnév]' or
                patternL=='[hónapnév] [időtartam]'):          # február hónap
            
            if '[időtartam]' in patternL:
                n,időtartam = outvaluesL
            else:
                n,=outvaluesL
                időtartam='hónap'

            if időtartam=='hónap':
                n=int(n)


                dtout=datetime(dt0.year,n,1)         # aktuális évben
                # Ha van kontextus, akkor eltolásra lehet szükség
                if tense=='future' and n<dt0.month: dtout=fn_dateadd(dtout,1,'year')
                elif tense=='past' and n>dt0.month: dtout=fn_dateadd(dtout,-1,'year')
                dtout2=fn_monthlastday(dtout)

        elif patternL=='[hétnapja]':
            hétnapja,=outvaluesL        # '1', ..., '7', 'hétvége'
            dtMonday=fn_monday(dt0)
            # Az aktuális héten
            if hétnapja=='hétvége': 
                dtout=fn_dateadd(dtMonday,5)    # szombat
                if tense=='past' and dtout>dt0: dtout=fn_dateadd(dtout,-7)
                dtout2=fn_dateadd(dtout,1)
            else:
                dtout=fn_dateadd(dtMonday,int(hétnapja)-1)
                # Ha van kontextus, akkor eltolásra lehet szükség
                if tense=='future' and dtout<dt0: dtout=fn_dateadd(dtout,7)
                elif tense=='past' and dtout>dt0: dtout=fn_dateadd(dtout,-7)

        elif patternL=='[évszak]':
            value,=outvaluesL
            
            year=dt0.year
            if value=='12' and dt0.month<6: year-=1    # a tél átnyúlik az évhatáron, ezért nem teljesen egyértelmű az "ezen a télen" ... jelentése

            dtout=datetime(year,int(value),1)    # kezdődátum
            dtout2=fn_monthlastday(fn_dateadd(dtout,2,'month'))

            # Ha van kontextus, akkor eltolásra lehet szükség
            if tense=='future' and dt0>dtout2:   # "télen hideg lesz",  az aktuális tél után vagyunk
                dtout=fn_dateadd(dtout,1,'year')
                dtout2=fn_monthlastday(fn_dateadd(dtout,2,'month'))
            elif tense=='past' and dt0<dtout:   # "télen hideg volt",  az aktuális tél előtt vagyunk
                dtout=fn_dateadd(dtout,-1,'year')
                dtout2=fn_monthlastday(fn_dateadd(dtout,2,'month'))
            
        elif (patternL=='[évnapja]' or              # karácsony, újév
                patternL=='[szám] [évnapja]' or
                patternL=='[dátum] [évnapja]' or
                patternL=='[múltjövő] [évnapja]'):       # 2023 karácsony
                
            year=None
            if patternL=='[évnapja]':              # karácsony, újév
                évnapja, = outvaluesL
                year=dt0.year
            elif patternL=='[szám] [évnapja]':
                szám,évnapja = outvaluesL
                year=szám
            elif patternL=='[dátum] [évnapja]':
                dátum,évnapja = outvaluesL
                dt1,dt2 = dátum
                if dt1.month!=1 or dt1.day!=1: return None    # csak teljes év jó
                year=dt1.year
            elif patternL=='[múltjövő] [évnapja]':
                múltjövő,évnapja = outvaluesL
                year=dt0.year
                if múltjövő=='múlt': year=year-1
                elif múltjövő=='jövő': year=year+1

            dtout,dtout2 = sub_évnapja(year,évnapja)

        elif patternL=='[múltjövő]':    # 'most'
            múltjövő, = outvaluesL
            if múltjövő=='most' and beginwith(invaluesL[0],'most'): dtout=dt0

        elif (patternL=='[forduló]' or              # "az ezredfordulón"   "az előző századvégen"
              patternL=='[múltjövő] [forduló]'):

            if patternL=='[forduló]':
                forduló, = outvaluesL
                múltjövő = 'most'
            elif patternL=='[múltjövő] [forduló]':
                múltjövő,forduló = outvaluesL

            if forduló in ['századvég','ezredvég']: dtout=datetime(1999,12,31)
            elif forduló in ['századelő','ezredelő']: dtout=datetime(2000,1,1)
                
            if múltjövő!='most':
                if beginwith(forduló,'század'): years=100
                elif beginwith(forduló,'ezred'): years=1000

                if múltjövő=='múlt': dtout=fn_dateadd(dtout,-years,'year')
                elif múltjövő=='jövő': dtout=fn_dateadd(dtout,years,'year')

        elif (patternL=='[egyedi]'):        # "rendszerváltás"    év,nap,időszak is lehet (korlátozottan használható mert a mintában pl nem lehet szám szó) 
            egyedi,=outvaluesL

            dtout,dtout2 = text2date(egyedi,dt0,outtype='first_tuple')



        elif patternL=='[szám] [időtartam] [múlvaezelőtt]':       # 2 nap múlva, 3 év múlva,  négy hét múlva,  egy évvel ezelőtt
            # FONTOS:  a "[szám] [időtartam]" mintázatnál van egy hasonló eset:   "3 éve",  "két hónapja",   "5 hete" 
            szám,időtartam,múlvaezelőtt=outvaluesL
            if múlvaezelőtt=='ezelőtt': szám=-szám
                
            if időtartam=='nap':
                dtout=fn_dateadd(dt0,szám)
            elif időtartam=='hét':
                dtout=fn_dateadd(dt0,7*szám)
            elif időtartam=='hónap':
                dtout=fn_dateadd(dt0,szám,'month')
            elif időtartam=='év':
                dtout=fn_dateadd(dt0,szám,'year')

        elif patternL=='[szám] [időtartam]':            # "2022 év", "90-es években", "két hete", "12 éve", "20 napja"
                                                        # XIX. században, második félévben, utolsó negyedévben, 12. héten, 120. napon
            szám,időtartam = outvaluesL

            szám_in = invaluesL[0]
            időtartam_in=invaluesL[1]
            #  "két hete",  "három hónapja",  "12 éve"   "20 napja"  (kétértelmű, de időpontot is jelenthet, "... ezelőtt" jelentéssel) 
            if időtartam_in in ['éve','honapja','hete','napja']: 
                szám=-szám
                if időtartam=='nap':
                    dtout=fn_dateadd(dt0,szám)
                elif időtartam=='hét':
                    dtout=fn_dateadd(dt0,7*szám)
                elif időtartam=='hónap':
                    dtout=fn_dateadd(dt0,szám,'month')
                elif időtartam=='év':
                    dtout=fn_dateadd(dt0,szám,'year')

            # "'90-es évek", "1870-es évek", "1200-as évek",
            elif beginwith(időtartam_in,'évek|évei') and (szám % 10)==0:
                if (szám % 100) ==0:
                    dtout=datetime(szám,1,1)
                    dtout2=datetime(szám+99,12,31)
                elif szám<100:      # jelen vagy múltbelinek tekintem ('30-as évek: 1930-as évek)
                    évszázad0 = dt0.year//100
                    évtized0 = (dt0.year - évszázad0*100)//10
                    évtized = szám//10
                    if évtized>évtized0: évszázad0-=1
                    dtout=datetime(évszázad0*100 + szám,1,1)
                    dtout2=datetime(évszázad0*100 + szám + 9,12,31)
                else:
                    dtout=datetime(szám,1,1)
                    dtout2=datetime(szám + 9,12,31)
               
            elif időtartam=='év' and szám>1900 and szám<=2100:      # ha nincs utána pont, akkor is tekintse konkrét évszámnak
                dtout=datetime(szám,1,1)
                dtout2=datetime(szám,12,31)

            elif időtartam=='évszázad':          # időszámítás kezdete óta
                if szám>=1 or szám<=30:
                    dtout=datetime((szám-1)*100,1,1)
                    dtout2=datetime((szám-1)*100 + 99,12,31)

            # [sorszám] [időtartam]
            #   Csak akkor jöjjön ide, ha előtte nem egyszavas dátumkifejezés áll
            #     (ebben az esetben először az egyszavas dátum lefordítása kell, majd utána a "[dátum] [szám] [időtartam]" mintázat
            elif sub_sorszam(szám_in) and not patternword_előtte in ['[szám]','[évszak]','[időtartam]','[utánielőtti]']:
                if időtartam=='évtized':       # aktuális évszázadban
                    if szám>=1 or szám<=10:
                        year=(dt0.year//100) * 100
                        dtout=datetime(year + (szám-1)*10,1,1)
                        dtout2=datetime(year + (szám-1)*10 + 9,12,31)
                elif időtartam=='év':            # időszámítás kezdete óta
                    if szám>=1 or szám<=2100:
                        dtout=datetime(szám,1,1)
                        dtout2=datetime(szám,12,31)
                elif időtartam=='félév':            # aktuális évben
                    if szám==1 or szám==2:
                        dtout=datetime(dt0.year,1+(szám-1)*6,1)
                        dtout2=fn_monthlastday(fn_dateadd(dtout,5,'month'))
                elif időtartam=='negyedév':         # aktuális évben
                    if szám>=1 and szám<=4:
                        dtout=datetime(dt0.year,1+(szám-1)*3,1)
                        dtout2=fn_monthlastday(fn_dateadd(dtout,2,'month'))
                elif időtartam=='hónap':            # aktuális évben
                    if szám>=1 and szám<=12:
                        dtout=datetime(dt0.year,szám,1)
                        dtout2=fn_monthlastday(dtout)
                elif időtartam=='hét':              # aktuális évben
                    if szám>=1 and szám<=53:
                        dtMonday=fn_monday(datetime(dt0.year,1,1))   # az év első hetének hétfői napja (általában az előző év végére esik)
                        dtout=fn_dateadd(dtMonday,(szám-1)*7,'day')
                        dtout2=fn_dateadd(dtout,6,'day')
                        if dtout.year<dt0.year: dtout=datetime(dt0.year,1,1)
                        if dtout2.year>dt0.year: dtout2=datetime(dt0.year,12,31)
                        if dtout>dtout2: dtout=None         # szám=53 esetén fordulhat elő
                elif időtartam=='nap':              # aktuális évben
                    if szám>=1 and szám<=366:
                        dtout=fn_dateadd(datetime(dt0.year,1,1),szám-1,'day')
                        if dtout.year>dt0.year: dtout=None


        elif (patternL=='[szám] [időtartam] [múlvaezelőtt] [hónapnév]' or       # 3 évvel ezelőtt decemberben, két év múlva októberben
                patternL=='[szám] [időtartam] [hónapnév]'):                       # 3 éve decemberben, két éve októberben
                
            if patternL=='[szám] [időtartam] [múlvaezelőtt] [hónapnév]':
                szám,időtartam,múlvaezelőtt,hónap=outvaluesL
                if múlvaezelőtt=='ezelőtt': szám=-szám
            elif patternL=='[szám] [időtartam] [hónapnév]':
                szám,időtartam,hónap=outvaluesL
                időtartam_in=invaluesL[1]
                if időtartam_in!='éve': időtartam=''
                szám=-szám
            if időtartam=='év':
                if int(hónap)>=1 and int(hónap)<=12:
                    dtout=datetime(dt0.year+szám,int(hónap),1)
                    dtout2=fn_monthlastday(dtout)

        elif (patternL=='[szám] [időtartam] [múlvaezelőtt] [szám]' or       # 2 hónappal ezelőtt 5-én, három hónap múlva hatodikán
                patternL=='[szám] [időtartam] [szám]'):                       # három hónapja negyedikén

            if patternL=='[szám] [időtartam] [múlvaezelőtt] [szám]':
                szám,időtartam,múlvaezelőtt,n=outvaluesL
                szám_in=invaluesL[3]
                if múlvaezelőtt=='ezelőtt': szám=-szám
            elif patternL=='[szám] [időtartam] [szám]':
                szám,időtartam,n=outvaluesL
                szám_in=invaluesL[2]
                időtartam_in=invaluesL[1]
                if időtartam_in!='honapja': időtartam=''
                szám=-szám

            if időtartam=='hónap':
                dt0=fn_dateadd(dt0,szám,'month')
                if n>=1 and n<=31:   # and (endwith(szám_in,'\.') or vanbenne(szám_in,'dika|dike|diká|diké|elsej')):
                    try: dtout=datetime(dt0.year,dt0.month,n)
                    except: dtout=None

        elif (patternL=='[szám] [időtartam] [múlvaezelőtt] [hétnapja]' or       # 4 héttel korábban pénteken, két hét múlva hétvégén
                patternL=='[szám] [időtartam] [hétnapja]'):        # két hete szombaton
                
            if patternL=='[szám] [időtartam] [múlvaezelőtt] [hétnapja]':
                szám,időtartam,múlvaezelőtt,hétnapja=outvaluesL
                if múlvaezelőtt=='ezelőtt': szám=-szám
            elif patternL=='[szám] [időtartam] [hétnapja]':
                szám,időtartam,hétnapja=outvaluesL
                szám=-szám
                időtartam_in=invaluesL[1]
                if időtartam_in!='hete': időtartam=''
                
            if időtartam=='hét':
                dt0=fn_dateadd(dt0,7*szám)
                dtMonday=fn_monday(dt0)
                if hétnapja=='hétvége': 
                    dtout=fn_dateadd(dtMonday,5)    # szombat
                    dtout2=fn_dateadd(dtout,1)
                else:
                    dtout=fn_dateadd(dtMonday,int(hétnapja)-1)


        elif (patternL=='[időtartam]'):             # "a héten" "a hónap", "az év", "a félév", "az évszázad"
            időtartam,=outvaluesL
            időtartam_in,=invaluesL
            if beginwith(időtartam_in,'a |az '):      # a keresőmintákban benne van a névelős változat is (pl. "a hét")
                dtout,dtout2 = sub_időtartam_to_időszak(időtartam,'most',dt0)

        elif patternL=='[múltjövő] [időtartam]':           # múlt héten, előző évben, következő hónapban, jövő félévben, következő héten
            múltjövő,időtartam=outvaluesL
            dtout,dtout2 = sub_időtartam_to_időszak(időtartam,múltjövő,dt0)

        elif (patternL=='[múltjövő] [évszak]' or                       # múlt nyáron, ezen a télen, jövő tavasszal
                patternL=='[múltjövő] [időtartam] [évszak]'):           # múlt év nyarán, jövő év tavasszal

            időtartam=''
            if '[időtartam]' in patternL:
                múltjövő,időtartam,évszak = outvaluesL
                if időtartam!='év': return None
            else:
                múltjövő,évszak = outvaluesL        # évszak: '3','6','9','12'
                # múltjövő pontosítása  (múlt versus előzős,   jövő versus következő)
                múltjövő_in=invaluesL[0]
                if múltjövő=='jövő' and (beginwith(múltjövő_in,'jövö')==''): múltjövő='következő'
                elif múltjövő=='múlt' and (beginwith(múltjövő_in,'mult')==''): múltjövő='előző'

            year=dt0.year
            month=dt0.month

            # A tél külön kezelendő, mert átfed az évváltással
            if évszak=='12':
                if múltjövő=='következő':
                    if month==12: year+=1       # decemberben már a következő évi tél
                elif múltjövő=='előző':
                    if month<3: year-=2         # januárfebruárban még az előző tél
                    else: year-=1               # egyébként az év eleji tél
                else:
                    if múltjövő =='most': year-=1
                    elif múltjövő =='múlt': year-=2
                    if month>=7: year+=1 # félévtől már a következő évire utal
            
            else:
                if múltjövő =='múlt': year-=1
                elif múltjövő =='jövő': year+=1
                elif múltjövő =='előző': 
                    if month<=int(évszak)+2: year-=1
                elif múltjövő =='következő': 
                    if month>=int(évszak): year+=1


            dtout=datetime(year,int(évszak),1)
            dtout2=fn_monthlastday(fn_dateadd(dtout,2,'month'))

        elif (patternL=='[múltjövő] [hónapnév]' or           # jövő májusban, múlt szeptemberben  (nem "következő"/"előző" a jelentése, hanem "jövőre","tavaly")
                patternL=='[múltjövő] [időtartam] [hónapnév]'):  # jövő év február
                
            időtartam=''
            if '[időtartam]' in patternL:
                múltjövő,időtartam,honap = outvaluesL
            else: 
                múltjövő,honap=outvaluesL      
            
            if időtartam=='' or időtartam=='év':
                year=dt0.year
                if múltjövő=='múlt': year=year-1
                elif múltjövő=='jövő': year=year+1

                n2=int(honap)
                dtout=datetime(year,n2,1)
                dtout2=fn_monthlastday(dtout)

        elif (patternL=='[múltjövő] [hónapnév] [szám]' or     # jővő május 3-án, előző február 4-én  (nem "következő"/"előző" a jelentése, hanem "jövőre","tavaly")
              patternL=='[múltjövő] [időtartam] [szám]'):     # jövő hónap harmadikán

            if patternL=='[múltjövő] [hónapnév] [szám]':
                múltjövő,honap,szám=outvaluesL   
                
                year=dt0.year
                if múltjövő=='múlt': year=year-1
                elif múltjövő=='jövő': year=year+1

            elif patternL=='[múltjövő] [időtartam] [szám]':
                múltjövő,időtartam,szám=outvaluesL
                if időtartam!= 'hónap':  return None
                honap=dt0.month
                year=dt0.year
            

            szám_in=invaluesL[2]
            if not endwith(szám_in,'dik|elsö'):
                try: 
                    dtout=datetime(year,int(honap),szám)
                    if patternL=='[múltjövő] [időtartam] [szám]':
                        if múltjövő=='múlt': dtout=fn_dateadd(dtout,-1,'month')
                        elif múltjövő=='jövő': dtout=fn_dateadd(dtout,1,'month')
                except: dtout=None



        elif (patternL=='[múltjövő] [hétnapja]' or                      # "jövő szombaton"
                patternL=='[múltjövő] [időtartam] [hétnapja]'):           # "múlt hét pénteken"
            időtartam=''
            if '[időtartam]' in patternL:
                múltjövő,időtartam,hétnapja = outvaluesL
            else: 
                múltjövő,hétnapja = outvaluesL

            if időtartam=='' or időtartam=='hét':
                dtMonday=fn_monday(dt0)
                if múltjövő=='múlt': dtMonday=fn_dateadd(dtMonday,-7) 
                elif múltjövő=='jövő': dtMonday=fn_dateadd(dtMonday,7) 

                if hétnapja=='hétvége': 
                    dtout=fn_dateadd(dtMonday,5)    # szombat
                    dtout2=fn_dateadd(dtout,1)
                else: dtout=fn_dateadd(dtMonday,int(hétnapja)-1)


        elif (patternL=='[dátum] [utánelőtt]'):
            dátum,utánelőtt = outvaluesL
            dt1,dt2 = dátum
            if dt2==None: dt2=dt1

            if utánelőtt=='előtt':
                dtout=datetime(1,1,1)
                dtout2=fn_dateadd(dt1,-1)
            elif utánelőtt=='után':
                dtout=fn_dateadd(dt2,1)
                dtout2=datetime(9999,12,31)



        elif (patternL=='[dátum] [utánelőtt] [szám] [időtartam]' or   # szombat után két nappal      január 10 előtt 2 héttel
                patternL=='[szám] [időtartam] [dátum] [utánelőtt]'):    # két nappal március után
                
            if patternL=='[dátum] [utánelőtt] [szám] [időtartam]':
                dátum,utánelőtt,szám,időtartam = outvaluesL
            elif patternL=='[szám] [időtartam] [dátum] [utánelőtt]':
                szám,időtartam,dátum,utánelőtt = outvaluesL
            
            dt1,dt2 = dátum
            if dt2==None: dt2=dt1

            if utánelőtt=='után':
                előjel=1
                dt0=dt2
            elif utánelőtt=='előtt':
                előjel=-1
                dt0=dt1

            szám=int(szám)

            if időtartam=='nap': dtout=fn_dateadd(dt0,szám*előjel)
            elif időtartam=='hét': dtout=fn_dateadd(dt0,7*szám*előjel)   
            elif időtartam=='hónap': dtout=fn_dateadd(dt0,szám*előjel,'month')
            elif időtartam=='negyedév':  dtout=fn_dateadd(dt0,3*szám*előjel,'month')
            elif időtartam=='félév': dtout=fn_dateadd(dt0,6*szám*előjel,'month')
            elif időtartam=='év': dtout=fn_dateadd(dt0,szám*előjel,'year')
            elif időtartam=='évtized':  dtout=fn_dateadd(dt0,szám*10*előjel,'year')
            elif időtartam=='évszázad':  dtout=fn_dateadd(dt0,szám*100*előjel,'year')
            elif időtartam=='évezred':  dtout=fn_dateadd(dtFirst,szám*1000*előjel,'year')

        elif (patternL=='[dátum] [utánielőtti] [szám] [időtartam]' or    # "szombat utáni első hét"  "január 10 előtti 2 nap"
                patternL=='[dátum] [utánielőtti] [időtartam]'):            # "jövő hét utáni hét"
            if '[szám]' in patternL:
                dátum,utánielőtti,szám,időtartam = outvaluesL
                szám_in = invaluesL[2]
            else:
                dátum,utánielőtti,időtartam = outvaluesL
                szám = 1
                szám_in=''

            dt1,dt2 = dátum
            if dt2==None: dt2=dt1

            if utánielőtti=='utáni':
                előjel=1
                dt0=dt2
            elif utánielőtti=='előtti':
                előjel=-1
                dt0=dt1

            szám=int(szám)
            # sorszám vagy normál szám?
            if sub_sorszam(szám_in):   # szombat utáni első nap,   tavasz előtti utolsó hét  (naptári hét, naptári hónap, ...)
                if időtartam=='nap': dtout=fn_dateadd(dt0,szám*előjel)
                elif időtartam=='hét':
                    dtMonday=fn_monday(dt0)             # az aktuális hét első napja
                    dtout=fn_dateadd(dtMonday,7*szám*előjel)   # előző, aktuális, következő
                    dtout2=fn_dateadd(dtout,6)          
                elif időtartam=='hónap':
                    dtMonthfirst=dt0.replace(day=1)
                    dtout=fn_dateadd(dtMonthfirst,szám*előjel,'month')
                    dtout2=fn_monthlastday(dtout)
                elif időtartam=='negyedév':
                    dtFirst=dt0.replace(month=(1 + ((dt0.month-1)//3) * 3))
                    dtFirst=dt0.replace(day=1)
                    dtout=fn_dateadd(dtFirst,3*szám*előjel,'month')
                    dtout2=fn_monthlastday(fn_dateadd(dtout,2,'month'))
                elif időtartam=='félév':
                    dtFirst=dt0.replace(month=(1 + ((dt0.month-1)//6) * 6))
                    dtFirst=dt0.replace(day=1)
                    dtout=fn_dateadd(dtFirst,6*szám*előjel,'month')
                    dtout2=fn_monthlastday(fn_dateadd(dtout,5,'month'))
                elif időtartam=='év':
                    dtFirst=datetime(dt0.year,1,1)
                    dtout=fn_dateadd(dtFirst,szám*előjel,'year')
                    dtout2=datetime(dtout.year,12,31)
                elif időtartam=='évtized':
                    dtFirst=datetime((dty.year//10)*10,1,1)
                    dtout=fn_dateadd(dtFirst,szám*10*előjel,'year')
                    dtout2=datetime(dtout.year+9,12,31)
                elif időtartam=='évszázad':
                    dtFirst=datetime((dty.year//100)*100,1,1)
                    dtout=fn_dateadd(dtFirst,szám*100*előjel,'year')
                    dtout2=datetime(dtout.year+99,12,31)
                elif időtartam=='évezred':
                    dtFirst=datetime((dty.year//1000)*1000,1,1)
                    dtout=fn_dateadd(dtFirst,szám*1000*előjel,'year')
                    dtout2=datetime(dtout.year+999,12,31)
            
            else:   # szombat utáni 2 nap      január 3 előtti 3 hét
                if időtartam=='nap': dt1=fn_dateadd(dt0,szám*előjel)
                elif időtartam=='hét': dt1=fn_dateadd(dt0,7*szám*előjel)   
                elif időtartam=='hónap': dt1=fn_dateadd(dt0,szám*előjel,'month')
                elif időtartam=='negyedév':  dt1=fn_dateadd(dt0,3*szám*előjel,'month')
                elif időtartam=='félév': dt1=fn_dateadd(dt0,6*szám*előjel,'month')
                elif időtartam=='év': dt1=fn_dateadd(dt0,szám*előjel,'year')
                elif időtartam=='évtized':  dt1=fn_dateadd(dt0,szám*10*előjel,'year')
                elif időtartam=='évszázad':  dt1=fn_dateadd(dt0,szám*100*előjel,'year')
                elif időtartam=='évezred':  dt1=fn_dateadd(dtFirst,szám*1000*előjel,'year')

                if dt1<dt0:
                    dtout=dt1
                    dtout2=fn_dateadd(dt0,-1)
                else:
                    dtout=fn_dateadd(dt0,1)
                    dtout2=dt1

        elif (patternL=='[dátum] [utánielőtti] [szám] [hétnapja]' or    # "december 1 utáni második hétvége"
                patternL=='[dátum] [utánielőtti] [hétnapja]'):
            if '[szám]' in patternL:
                dátum,utánielőtti,szám,hétnapja = outvaluesL
                szám_in = invaluesL[2]
            else:
                dátum,utánielőtti,hétnapja = outvaluesL
                szám = 1
                szám_in='elsö'

            dt1,dt2 = dátum
            if dt2==None: dt2=dt1

            szám=int(szám)
            if sub_sorszam(szám_in):   # szombat utáni első hétvége,   karácsony előtti második hétvége
                length=1
                if hétnapja=='hétvége':
                    length=2
                    hétnapja='6'
                if utánielőtti=='utáni':
                    dt0=fn_dateadd(dt2,1)
                    # Az első megfelelő nap az időszakban
                    dtMonday=fn_monday(dt0)
                    dtFirst=fn_dateadd(dtMonday,int(hétnapja)-1)
                    if dtFirst<dt0: dtFirst=fn_dateadd(dtFirst,7)
                    dtout = fn_dateadd(dtFirst,(szám-1)*7)
                    if length==2: dtout2=fn_dateadd(dtout,1)
                elif utánielőtti=='előtti':
                    dt0=fn_dateadd(dt1,-1)
                    # Az első megfelelő nap az időszakban
                    dtMonday=fn_monday(dt0)
                    dtFirst=fn_dateadd(dtMonday,int(hétnapja)-1)
                    if dtFirst>dt0: dtFirst=fn_dateadd(dtFirst,-7)
                    dtout=fn_dateadd(dtFirst,-(szám-1)*7)
                    if length==2: dtout2=fn_dateadd(dtout,1)

        elif (patternL=='[dátum] [elejevége] [elejevége]' or      #  "január közepétől a végéig"
                patternL=='[időtartam] [elejevége] [elejevége]'):   #  "a hét közepétől a végéig"
                
            bOk=True
            if 'időtartam' in patternL:
                if patternword_előtte=='[szám]': bOk=False   # nem állhat előtte szám, pl. "XIX. század közepétől a végéig"
                else:
                    időtartam = outvaluesL[0]
                    outvaluesL[0]=sub_időtartam_to_időszak(időtartam,'most',dt0)
                    patternL=patternL.replace('időtartam','dátum')
                
            if bOk:
                dátum,elejevége,elejevége2 = outvaluesL
                szám1=int(elejevége)
                szám2=int(elejevége2)
                if szám2>szám1:
                    dt1,dt2 = dátum
                    if dt2!=None and dt2>dt1:       # időszak szerepel a dátumban
                        # A "közepe" itt nem a középső harmadra, hanem a középső napra utal
                        dtkozep=fn_dateadd(dt1,fn_daydiff(dt2,dt1)//2)
                        if szám1==1: dtout=dt1
                        elif szám1==2: dtout=dtkozep
                        if szám2==2: dtout2=dtkozep
                        elif szám2==3: dtout2=dt2

        elif (patternL=='[dátum] [szám] [törtrész]' or    # "2022. első felében", "XIX. század második felében", "január harmadik hamadában",  "múlt hét első felében"
                patternL=='[dátum] [elejevége]' or         # "2022. elején","február közepén","XIX. század végén", "jövő hét elején"   
                patternL=='[időtartam] [szám] [törtrész]' or    # "az év első felében",   "a hét második felében"
                patternL=='[időtartam] [elejevége]'):      # "hét közepén", "a hónap elejétől", "év végén"

            bOk=True
            if 'időtartam' in patternL:
                if patternword_előtte=='[szám]': bOk=False   # nem állhat előtte szám, pl. "XIX. század első felében" példamondat "század első felében"-ként értelmeződne
                else:
                    időtartam = outvaluesL[0]
                    outvaluesL[0]=sub_időtartam_to_időszak(időtartam,'most',dt0)
                    patternL=patternL.replace('időtartam','dátum')
                
            if bOk:
                dátum=None
                if patternL=='[dátum] [szám] [törtrész]':
                    dátum,szám,törtrész = outvaluesL
                    szám_in = invaluesL[1]
                    if not sub_sorszam(szám_in): dátum=None

                elif patternL=='[dátum] [elejevége]':
                    dátum,elejevége = outvaluesL

                    # Ha a [dátum] végén is elejevége volt, akkor egyedi kezelés (pl. "január közepétől a végéig")
                    dátum_minta=invaluesL[0]
                    if endwith(dátum_minta,'\[elejevége\]'): dátum=None
                    else:
                        törtrész=3
                        szám=int(elejevége)         # Hányadik harmad:   eleje:1  közepe: 2  vége:3

                if dátum!=None:
                    dt1,dt2 = dátum
                    if dt2!=None and dt2>dt1:       # időszak szerepel a dátumban
                        oszto=int(törtrész)
                        if szám>=1 and szám<=oszto:
                            unit=fn_daydiff(dt2,dt1)/oszto
                            if szám==oszto:
                                dtout=fn_dateadd(dt2,-fn_round(unit),'day')
                                dtout2=dt2
                            else:
                                dtout=fn_dateadd(dt1,fn_round(unit*(szám-1)),'day')
                                dtout2=fn_dateadd(dtout,fn_round(unit),'day')
                

        elif (patternL=='[dátum] [szám] [időtartam]' or     # "jövő év második félévében", "XIX. század harmadik évtizedében", "február első hetében"
                patternL=='[dátum] [utolsó] [időtartam]'):    # "jövő év utolsó napján", "XIX. század utolsó évtizedében", "február első hetében"

            bUtolso = vanbenne(patternL,'utolsó')
            patternL=patternL.replace('utolsó','szám')

            dt1=None
            dátum,szám,időtartam = outvaluesL
            dt1,dt2 = dátum

            if dt1!=None and dt2!=None and dt2>dt1:       # időszak szerepel a dátumban
                szám_in = invaluesL[1]
                időtartam_in=invaluesL[2]
                if sub_sorszam(szám_in) or bUtolso:
                    if időtartam=='évtized' and (dt1.year % 10)==0 and dt1.month==1 and dt1.day==1:            
                        if bUtolso: szám=(dt2.year-dt1.year)//10 + 1
                        dtout=datetime(dt1.year + 10*(szám-1),1,1)
                        dtout2=datetime(dt1.year + 10*(szám-1) + 9,12,31)
                    elif időtartam=='év' and dt1.month==1 and dt1.day==1:            
                        if bUtolso: szám=(dt2.year-dt1.year)+1
                        dtout=datetime(dt1.year + (szám-1),1,1)
                        dtout2=datetime(dt1.year + (szám-1),12,31)
                    elif időtartam=='félév' and dt1.month==1 and dt1.day==1:            
                        if bUtolso: szám=(dt2.month-dt1.month)//6 + 1
                        dtout=datetime(dt1.year,1+(szám-1)*6,1)
                        dtout2=fn_monthlastday(fn_dateadd(dtout,5,'month'))
                    elif időtartam=='negyedév' and dt1.month==1 and dt1.day==1:         
                        if bUtolso: szám=(dt2.month-dt1.month)//3 + 1
                        dtout=datetime(dt1.year,1+(szám-1)*3,1)
                        dtout2=fn_monthlastday(fn_dateadd(dtout,2,'month'))
                    elif időtartam=='hónap' and dt1.day==1:            
                        if bUtolso: szám=(dt2.month-dt1.month) + 1
                        dtout=datetime(dt1.year,dt1.month + (szám-1),1)
                        dtout2=fn_monthlastday(dtout)
                    elif időtartam=='hét':              
                        dtMonday=fn_monday(dt1)   # az időszak első hetének hétfői napja (általában korábra esik az időszak első napjánál)
                        if bUtolso: szám=fn_daydiff(dt2,dtMonday)//7 + 1
                        dtout=fn_dateadd(dtMonday,(szám-1)*7,'day')
                        dtout2=fn_dateadd(dtout,6,'day')
                    elif időtartam=='nap':              
                        if bUtolso: szám=fn_daydiff(dt2,dt1) + 1
                        dtout=fn_dateadd(dt1,szám-1,'day')
                        dtout2=dtout

                    if dtout!=None:
                        if dtout>dt2: dtout=None     # a kezdődátum nem lehet későbbi az időszaknál
                        else: 
                            if dtout<dt1: dtout=dt1
                            if dtout2>dt2: dtout2=dt2
                            if dtout2<dtout: dtout=None
                    
                elif beginwith(időtartam_in,'évei|évek'):       # "XIX. század 20-as éveiben"
                    if dt1.year%100==0 and dt1.month==1 and dt1.day==1 and szám%10==0:
                        dtout=datetime(dt1.year + szám,1,1)
                        dtout2=datetime(dt1.year + szám + 9,12,31)

           
        elif (patternL=='[dátum] [szám] [hétnapja]' or     # jövő hónap első hétfőjén"
                patternL=='[dátum] [utolsó] [hétnapja]'):   # "február utolsó hétvégéjén"

            bUtolso = vanbenne(patternL,'utolsó')
            patternL=patternL.replace('utolsó','szám')

            dt1=None
            if patternL=='[dátum] [szám] [hétnapja]':
                dátum,szám,hétnapja = outvaluesL
                dt1,dt2 = dátum
            elif patternL=='[szám] [szám] [hétnapja]':
                évszám,szám,hétnapja = outvaluesL
                if évszám>1000 and évszám<2100:
                    dt1=datetime(évszám,1,1)
                    dt2=datetime(évszám,12,31)
            elif patternL=='[hónapnév] [szám] [hétnapja]':
                hónap,szám,hétnapja = outvaluesL
                dt1=datetime(dt0.year,int(hónap),1)
                dt2=fn_monthlastday(dt1)

            if dt1!=None and dt2!=None and dt2>dt1:       # időszak szerepel a dátumban
                szám_in = invaluesL[1]
                if sub_sorszam(szám_in) or bUtolso:
                    length=1
                    if hétnapja=='hétvége':
                        length=2
                        hétnapja='6'
                    # Az első megfelelő nap az időszakban
                    dtFirst = fn_dateadd(dt1, (7 + int(hétnapja) - (dt1.weekday()+1)) % 7)
                    if bUtolso: szám=fn_daydiff(dt2,dtFirst)//7 + 1

                    dtout = fn_dateadd(dtFirst,(szám-1)*7)
                    dtout2=fn_dateadd(dtout,length-1)

                    if dtout!=None:
                        if dtout>dt2: dtout=None     # a kezdődátum nem lehet későbbi az időszaknál
                        else: 
                            if dtout<dt1: dtout=dt1
                            if dtout2>dt2: dtout2=dt2
                            if dtout2<dtout: dtout=None

        elif (patternL=='[dátum] [szám]'):             # "2023 februártól 5-től 10-ig",  "2023.02.05-10"
            dátum,szám = outvaluesL
                
            dt1,dt2 = dátum
            if dt2==None or dt2==dt1:
                dátum_minta = invaluesL[0]
                if endwith(dátum_minta,r'\[szám\]') and szám>=1 and szám<=31:
                    try: dtout2=datetime(dt1.year,dt1.month,szám)
                    except: dtout2=None
                    if dtout2!=None: dtout=dt1

            
        elif patternL=='[dátum] [dátum]':             # "2023 évben május végén",  "2023 februártól májusig",  "téltől tavaszig", "2023.02.01 - 2025.03.31"
            # Csak akkor jöjjön ide, ha után a nem [elejevége] áll
            #   Ha [elejevége] áll utána, akkor először a "[dátum] [elejevége]" mintázat érvényesüljön, és csak utána a "[dátum] [dátum]"
            if not patternword_utána in ['[elejevége]']:
                dátumA,dátumB = outvaluesL
                dt1A,dt2A = dátumA
                if dt2A==None: dt2A=dt1A

                dt1B,dt2B = dátumB
                if dt2B==None: dt2B=dt1B

                # Speciális esetek
                mintaA,mintaB = invaluesL
                # "2020 februártól augusztusig",  "2020 február-augusztus"
                if endwith(mintaA,'\[hónapnév\]') and beginwith(mintaB,'\[hónapnév\]'):
                    dt1B=dt1B.replace(year=dt1A.year)
                    dt2B=dt2B.replace(year=dt1A.year)
                # "2020 tavasztól őszig"
                elif endwith(mintaA,'\[évszak\]') and beginwith(mintaB,'\[évszak\]'):
                    dt1B=dt1B.replace(year=dt1A.year)
                    dt2B=dt2B.replace(year=dt1A.year)
                # "2020 második negyedévtől a harmadik negyedévig"
                if mintaA=='[dátum] [szám] [időtartam]' and mintaB=='[szám] [időtartam]':
                    dt1B=dt1B.replace(year=dt1A.year)
                    dt2B=dt2B.replace(year=dt1A.year)

                if dt1A<=dt2B and dt2A<=dt1B:
                    dtout=dt1A
                    dtout2=dt2B


 


        if dtout==None: return None
        else: return dtout,dtout2


    def sub_findpattern(pattern,invalues,outvalues,dt0):
        # ciklus az ablakméretre:  összes szó, összes-1 szó, összes-2 szó, ...
        patternwords=pattern.split()
        nWords=len(patternwords)

        # invalues és outvalues összehangolása a patternwords-szel (nem azonos az elemszámuk)
        invalues_sync=[None]*nWords
        outvalues_sync=[None]*nWords
        i_value=0
        for i,patternword in enumerate(patternwords):
            if patternword[0]=='[':
                invalues_sync[i]=invalues[i_value]
                outvalues_sync[i]=outvalues[i_value]
                i_value+=1

        # Ablak-pásztázási ciklus
        nWindowLenMax=6        # ennél több szóból álló minták nem érvényesülnek (lásd sub patterns)
        if nWindowLenMax>nWords: nWindowLenMax=nWords

        for nWindowLen in reversed(range(nWindowLenMax+1)):   
            for nWindowStart in range(nWords-nWindowLen+1):
                # patternL,invaluesL,outvaluesL beállítása
                patternL=' '.join(patternwords[nWindowStart:nWindowStart+nWindowLen])
                patternword_előtte=''
                if nWindowStart>0: patternword_előtte=patternwords[nWindowStart-1]
                patternword_utána=''
                if nWindowStart+nWindowLen<nWords: patternword_utána=patternwords[nWindowStart+nWindowLen]
                invaluesL=[]
                for value in invalues_sync[nWindowStart:nWindowStart+nWindowLen]:
                    if value: invaluesL.append(value)    # a None értékek nem kellenek
                outvaluesL=[]
                for value in outvalues_sync[nWindowStart:nWindowStart+nWindowLen]:
                    if value: outvaluesL.append(value)   # a None értékek nem kellenek
                # beazonosítható-e dátumként
                dates=sub_patterns(patternL,invaluesL,outvaluesL,dt0,patternword_előtte,patternword_utána)
                if dates:
                    pattern=' '.join([*patternwords[:nWindowStart],'[dátum]',*patternwords[nWindowStart+nWindowLen:]])
                    pattern=trim(pattern)
                    invalues_sync=[*invalues_sync[:nWindowStart],patternL,*invalues_sync[nWindowStart+nWindowLen:]]
                    outvalues_sync=[*outvalues_sync[:nWindowStart],dates,*outvalues_sync[nWindowStart+nWindowLen:]]
                    invalues=[]
                    for value in invalues_sync:
                        if value: invalues.append(value)    # a None értékek nem kellenek
                    outvalues=[]
                    for value in outvalues_sync:
                        if value: outvalues.append(value)   # a None értékek nem kellenek
                    return (pattern,invalues,outvalues)
        return None
    



    # Annnotálás
    pattern,invalues,outvalues = d_annotate(text,lookup_text2dateG,bNum=True,max_words_in_samples=2)
    # - a max_words_in_samples a ténylegesen alkalmazott keresőmintákhoz igazítandó (lookup_text2dateG)

    
    while True:
        result=sub_findpattern(pattern,invalues,outvalues,dt0)
        # Ha nincs további beazonosítás, akkor vége
        if result==None:
            # Megnézendő, hogy volt-e sikeres dátumbeazonosítás (az első lesz a return)
            dátumok=[]
            nValueIndex=-1
            patternwords=pattern.split()
            for patternword in patternwords:
                if patternword[0]=='[': nValueIndex+=1
                if patternword=='[dátum]':
                    date1,date2 = outvalues[nValueIndex]
                    if outtype=='first_tuple':
                        return (date1,date2)
                    else:
                        if date2 and date2!=date1: result=date1.strftime('%Y.%m.%d') + '-' + date2.strftime('%Y.%m.%d')
                        else: result=date1.strftime('%Y.%m.%d')
                        if outtype=='first+': result=(result + '\n' +
                            '   pattern: ' + pattern + '\n' +
                            '   outvalues: ' + str(outvalues) + '\n' +
                            '   invalues: ' + str(invalues))
                        if outtype in ['first','first+']:
                            return result
                    dátumok.append(result)
            if len(dátumok)>0:
                return ','.join(dátumok)

            # sikertelen  (egyetlen dátum sem volt a szövegben)
            if outtype=='first':  result=''
            elif outtype=='first+':  result='sikertelen     pattern: ' + pattern + '  outvalues: ' + str(outvalues)
            elif outtype=='first_tuple': result=(None,None)
            return result

        # Ha volt beazonosítás, akkor a pásztázási ciklus újrakezdése
        else:
            pattern,invalues,outvalues = result
            # Ha a teljes szöveget sikerült dátumként beazonosítani
            if pattern=='[dátum]': 
                date1,date2 = outvalues[0]
                if outtype=='first_tuple':
                    result = (date1,date2)
                else:
                    if date2 and date2!=date1: result=date1.strftime('%Y.%m.%d') + '-' + date2.strftime('%Y.%m.%d')
                    else: result=date1.strftime('%Y.%m.%d')
                    if outtype=='first+': result=(result + '\n' +
                        '   pattern: ' + pattern + '\n' +
                        '   outvalues: ' + str(outvalues) + '\n' +
                        '   invalues: ' + str(invalues))
                return result
            # egyébként újrakezdi a ciklust





# KERESŐMINTÁK (lookup)
# Formátum soronként:   keresőminták:entityname,outvalue
#      - keresőminták:  több minta is megadható '|' határolással
#         - "." a keresőminta végén, ha teljes szavas találat az elvárás (egyébként beginwith)
#         - egy-egy keresőminta több szavas is lehet (pl. "mindenszentek nap"). Többszavas minta esetén számít a szavak sorrendje. 
#         - a szavakon belül csak betűk és számjegyek lehetnek (de a "hét" szón kívül nem lehet benne szám-szó, mert a számokat külön kezeli az algoritmus) 
#         - kisbetű-nagybetű és hosszú ékezet eltérés érdektelen (a-á, e-é, o-ö, u-ü eltérés viszont számít)
#      - entityname:  szögletes zárójeles helyettesítőjelként kerül be a mintázatba  (kötelezően egyszavas)
#         - az entityname lehet "none": ebben az esetben nem a helyettesítőjel íródik be a mintázatba, hanem az outvalue 
#           (szabványosítás, szinonimák összevonása)
#         - új entity-név bevezetése esetén a text2date függvény sub_patterns függvényében fel kell venni az entity-névre támaszkodó
#           mintázatokat, és a hozzájuk tartozó interpretáló algortimust
#      - outvalue:   a szabványosított / kivonatolt érték   (a mintázatot értelmező algoritmus támaszkodik rá)
#         - "évnapja" esetén speciális kódolás:  
#             "MMdd",  "MMdd:MMdd",  "húsvét-1", "húsvét-2:húsvét+1  (a húsvét helyébe húsvét vasárnap dátuma íródik)

# Speciális entity:  "stopword"    Kimarad a mintázatból és az outvalues-ből is.

lookup_text2dateG={
    'hétfő|h.':'hétnapja,1',
    'kedd|k.':'hétnapja,2',
    'szerd|sze.|szer.':'hétnapja,3',
    'csüt|cs.':'hétnapja,4',
    'pént|p.':'hétnapja,5',
    'szomb|szo.':'hétnapja,6',
    'vasár|vas.|v.':'hétnapja,7',
    'hétvég':'hétnapja,hétvége',

    'január|jan.':'hónapnév,1',
    'február|febr.':'hónapnév,2',
    'március|márc.':'hónapnév,3',
    'április|ápr.':'hónapnév,4',
    'május|máj.':'hónapnév,5',
    'június|jún.|júni.':'hónapnév,6',
    'július|júl.|júli.':'hónapnév,7',
    'augusztus|aug.':'hónapnév,8',
    'szeptember|szept.':'hónapnév,9',
    'október|okt.':'hónapnév,10',
    'november|nov.':'hónapnév,11',
    'december|dec.':'hónapnév,12',

    'ma.|mai.|má':'maholnap,0',
    'holnap':'maholnap,1',
    'holnapután':'maholnap,2',
    'tegnap':'maholnap,-1',
    'tegnapelőtt':'maholnap,-2',

    'századvég|a századvég':'forduló,századvég',
    'századelő|a századelő|századforduló|a századforduló|évszázad forduló|az évszázad forduló':'forduló,századelő',
    'ezredvég|az ezredvég':'forduló,ezredvég',
    'ezredelő|az ezredelő|ezredforduló|az ezredforduló|évezred forduló|az évezred forduló':'forduló,ezredelő',

    'idén':'szám,' + str(datetime.now().year),
    'jövőre':'szám,' + str(datetime.now().year+1),
    'tavaly':'szám,' + str(datetime.now().year-1),
    'tavalyelőtt':'szám,' + str(datetime.now().year-2),

    'évszázad|század|szd.|a század|az évszázad':'időtartam,évszázad',
    'évtized|az évtized':'időtartam,évtized',
    'év.|éve.|évek|évei|évtől|évig|évben|évvel|évé|évre|az év':'időtartam,év',
    'félév|a félév':'időtartam,félév',
    'negyedév|a negyedév':'időtartam,negyedév',
    'hónap|hó.|a hónap':'időtartam,hónap',
    'hét|hete.|heté|a hét':'időtartam,hét',
    'nap':'időtartam,nap',

    'tavasz|tavasszal|a tavasz':'évszak,3',
    'nyár|nyarán|a nyár':'évszak,6',
    'ősz|ősszel|az ősz':'évszak,9',
    'tél|telén|a tél':'évszak,12',

    'múlt|elmúlt|előző|legutóbbi':'múltjövő,múlt',
    'jövő|következő|legközelebbi':'múltjövő,jövő',
    'most|jelen|ezen|idei|aktuális|e.|ebben|ettől|eddig':'múltjövő,most',

    'múlva|ezután|azután|később|rá.|követően':'múlvaezelőtt,múlva',
    'ezelőtt|azelőtt|korábban|megelőzően':'múlvaezelőtt,ezelőtt',

    'után':'utánelőtt,után',
    'előtt':'utánelőtt,előtt',

    'utáni|követő|kezdve':'utánielőtti,utáni',
    'előtti|megelőző':'utánielőtti,előtti',

    'utolsó': 'utolsó,utolsó',
    'utolsó előtti':'utolsó,utolsóelőtti',

    'fele.|felében|felétől|feléig':'törtrész,2',
    'harmada|harmadá':'törtrész,3',
    'negyede|negyedé':'törtrész,4',

    'elej|kezdőnap|kezdet':'elejevége,1',
    'közepe.|közepé':'elejevége,2',
    'vége.|végé|zárónap|zárás':'elejevége,3',

    'karácsony':'évnapja,1225:1226',
    'szenteste|karácsony este':'évnapja,1224',
    'szilveszter|évvég':'évnapja,1231',
    'újév|évkezdet':'évnapja,0101',
    'mikulás|télapó':'évnapja,1206',
    'húsvét':'évnapja,húsvét_',
    'húsvétpéntek|húsvét péntek':'évnapja,húsvét-2',
    'húsvétszombat|húsvét szombat':'évnapja,húsvét-1',
    'húsvétvasárnap|húsvét vasárnap':'évnapja,húsvét',
    'húsvéthétfő|húsvét hétfő':'évnapja,húsvét+1',
    'pünkösd':'évnapja,pünkösd_',
    'pünkösdvasárnap|pünkösd vasárnap':'évnapja,pünkösd',
    'pünkösdhétfő|pünkösd hétfő':'évnapja,pünkösd+1',
    'halottak nap':'évnapja,1102',
    'mindenszentek|mindenszentek nap':'évnapja,1101',
    'államalapítás nap|államalapítás ünnep|alkotmány nap': 'évnapja,0820',
    'március idus':'évnapja,0315',

    'rendszerváltás':'egyedi,1990',         # írásjelek és szám-szavak nem szerepelhetnek a mintákban (korlázott az egyedi mintázatok felvételi lehetősége)



    'a.|az.|és.|ezt.|azt.|tól.|től.|ig.|áig.|éig.|i.|ai.|ei.|án.|én.|jén.|jei.|ji.|tól.|től.|as.|es.':'stopword'
    }


